#ifndef	MVLIB_H_INLUDED
#define	MVLIB_H_INLUDED

#include <math.h>
#include <malloc.h>
#include <stdlib.h>
#include <assert.h>

#define	VEC(v)	((vec_t)(v))

#ifdef  M_PI
#undef  M_PI
#endif
#define M_PI                            VEC(3.14159265358979323846264338328)
#ifdef  M_2PI
#undef  M_2PI
#endif
#define M_2PI                           VEC(6.28318530717958647692528676656)
#define M_PI2                           VEC(1.57079632679489661923132169164)
#define M_PI3                           VEC(1.04719755119659774615421446109)
#define M_PI4                           VEC(.785398163397448309615660845820)
#define M_PI6                           VEC(.523598775598298873077107230548)

/*  Angles in radians; after underscore - angle in degrees              */
#define ANGLE_0                         VEC(0.0)
#define ANGLE_1                         VEC(.0174532925199432957692369076849)
#define ANGLE_5                         VEC(.0872664625997164788461845384245)
#define ANGLE_10                        VEC(.174532925199432957692369076849)
#define ANGLE_15                        VEC(.261799387799149436538553615273)
#define ANGLE_20                        VEC(.349065850398865915384738153697)
#define ANGLE_30                        M_PI6
#define ANGLE_45                        M_PI4
#define ANGLE_60                        M_PI3
#define ANGLE_72                        VEC(1.25663706143591729538505735331)
#define ANGLE_90                        M_PI2
#define ANGLE_120                       VEC(2.09439510239319549230842892219)
#define ANGLE_135                       VEC(2.35619449019234492884698253746)
#define ANGLE_144                       VEC(2.51327412287183459077011470662)
#define ANGLE_180                       M_PI
#define ANGLE_225                       VEC(3.92699081698724154807830422910)
#define ANGLE_270                       VEC(4.71238898038468985769396507492)
#define ANGLE_315                       VEC(5.49778714378213816730962592074)
#define ANGLE_360                       M_2PI

typedef	unsigned	u32_t;
typedef	int			s32_t;

typedef	double	vec_t;
typedef	vec_t	vec2_t[2];
typedef	vec_t	vec3_t[3];
typedef vec_t	matrix3_t[3][3];
typedef vec_t	matrix4_t[4][4];
typedef vec_t	matrix3x4_t[3][4];

typedef struct	plane3_s
{
	vec3_t  normal;
	vec_t   dist;
} plane3_t, *plane3_p;

typedef struct	line3_s
{
	vec3_t origin;
	vec3_t dir;
} line3_t, *line3_p;

#define	vboolean	int

#ifndef	__cplusplus
#ifndef true
#define	false	0
#define	true	(!false)
#endif	/* !true */
#endif	/* __cplusplus */

#define	V_ASSERT(obj)	assert(obj)

#define	V_GLOBAL
#define	V_LOCAL
#define	memAlloc(len)			malloc(len)
#define	memFree(ptr)			{V_ASSERT(ptr);free(ptr);}
#define	memZero(ptr,len)		memset(ptr,0,len);
#define	memCopy(dst,src,len)	memcpy(dst,src,len);

#define	memZero32(ptr,len)		memset(ptr,0,(len) << 2);
#define	memCopy32(dst,src,len)	memcpy(dst,src,(len) << 2);

#define	MAXVEC	(1E+38)
#define	FLOAT_ZERO_EPSILON	(1E-8)

#define	v_abs(v)	(vec_t)fabs(v)
#define	IsZero(v)	(v_abs(v) < FLOAT_ZERO_EPSILON)
#define	v_fsincos(s,c,r)	*((vec_t*)s)=sin(r);*((vec_t*)c)=cos(r);
/*
 * Vector operations
 */
#define	vec2Zero(vec) \
	vec2Set(vec,0,0)

#define	vec2Set(vec,x,y) \
{ \
	(vec)[0] = (x); \
	(vec)[1] = (y); \
}

#define	vec2Copy(dst,src) \
{ \
	(dst)[0] = (src)[0]; \
	(dst)[1] = (src)[1]; \
}

#define	vec3Zero(dst) \
{ \
	(dst)[0] = 0; \
	(dst)[1] = 0; \
	(dst)[2] = 0; \
}

#define	vec3Set(dst, x, y, z) \
{ \
	(dst)[0] = x; \
	(dst)[1] = y; \
	(dst)[2] = z; \
}

#define	vec3Copy(dst, src) \
{ \
	(dst)[0] = (src)[0]; \
	(dst)[1] = (src)[1]; \
	(dst)[2] = (src)[2]; \
}

#define	vec2Add(dst,v1,v2) \
{ \
	(dst)[0] = (v1)[0] + (v2)[0]; \
	(dst)[1] = (v1)[1] + (v2)[1]; \
}

#define	vec3Add(dst, v1, v2) \
{ \
	(dst)[0] = (v1)[0] + (v2)[0]; \
	(dst)[1] = (v1)[1] + (v2)[1]; \
	(dst)[2] = (v1)[2] + (v2)[2]; \
}

#define	vec2Sub(dst,v1,v2) \
{ \
	(dst)[0] = (v1)[0] - (v2)[0]; \
	(dst)[1] = (v1)[1] - (v2)[1]; \
}

#define	vec3Sub(dst, v1, v2) \
{ \
	(dst)[0] = (v1)[0] - (v2)[0]; \
	(dst)[1] = (v1)[1] - (v2)[1]; \
	(dst)[2] = (v1)[2] - (v2)[2]; \
}

#define	vec2Scale(dst, v, scale) \
{ \
	vec_t local_scale = (vec_t)(scale); \
	(dst)[0] = (v)[0] * local_scale; \
	(dst)[1] = (v)[1] * local_scale; \
}

#define	vec3Scale(dst, v, scale) \
{ \
	vec_t local_scale = (vec_t)(scale); \
	(dst)[0] = (v)[0] * local_scale; \
	(dst)[1] = (v)[1] * local_scale; \
	(dst)[2] = (v)[2] * local_scale; \
}

#define	vec2Negate(dst,src) \
{ \
	(dst)[0] = -(src)[0]; \
	(dst)[1] = -(src)[1]; \
}

#define	vec3Negate(dst,src) \
{ \
	(dst)[0] = -(src)[0]; \
	(dst)[1] = -(src)[1]; \
	(dst)[2] = -(src)[2]; \
}

#define	vec3Dot(v1, v2) \
	((v1)[0] * (v2)[0] + (v1)[1] * (v2)[1] + (v1)[2] * (v2)[2])

#define	vec3Abs(v) \
	((vec_t)sqrt(vec3Dot(v,v)))

#define	vec3Unit(dst,src) \
{ \
	vec_t rlen; \
	rlen = vec3Abs(src); \
	if(rlen < FLOAT_ZERO_EPSILON) \
	{ \
		vec3Set(dst,0,0,1); \
	} \
	else \
	{ \
		rlen = (vec_t)1.0 / rlen; \
		vec3Scale(dst,src,rlen); \
	} \
}


/* Cross product */
#define	vec3Cross(dst,v1,v2) \
{ \
	(dst)[0] = (v1)[1] * (v2)[2] - (v1)[2] * (v2)[1]; \
	(dst)[1] = (v1)[2] * (v2)[0] - (v1)[0] * (v2)[2]; \
	(dst)[2] = (v1)[0] * (v2)[1] - (v1)[1] * (v2)[0]; \
}

/*
 *	vec3RotateAround - rotation of vector 'pt' around arbitrary 'dir'
 */
#define	vec3RotateAround(/*vec3_t*/dst,/*vec3_t*/dir,/*vec3_t*/point,/*vec_t*/radians) \
{ \
	vec3_t nd; \
	matrix3_t rot; \
	vec3Unit(nd,dir); \
	mat3Rotate(rot,nd,radians); \
	mat3VecMul(dst,rot,point); \
} \

/*
 * Matrix operations
 */

#define	mat3Zero(mat) \
	memZero(mat,9  *  sizeof(vec_t));

#define	mat4Zero(mat)\
	memZero(mat,16 * sizeof(vec_t));

#define	mat3x4Zero(mat)\
	memZero(mat,12 * sizeof(vec_t));

#define	mat3Copy(dst,src) \
	memCopy32((dst),src,9 * (sizeof(vec_t)>>2));

#define	mat4Copy(dst,src) \
	memCopy32(dst,src,sizeof(vec_t)<<2);

#define	mat3x4Copy(dst,src) \
	memCopy32(dst,src,3 * sizeof(vec_t));

#define	mat3Unit(mat) \
{ \
	mat3Zero(mat); \
	(mat)[0][0] = (mat)[1][1] = (mat)[2][2] = 1.0; \
}

#define	mat4Unit(mat) \
{ \
	mat4Zero(mat); \
	(mat)[0][0] = (mat)[1][1] = (mat)[2][2] = (mat)[3][3] = 1.0; \
}

#define	mat3x4Unit(mat) \
{ \
	mat3x4Zero(mat); \
	(mat)[0][0] = (mat)[1][1] = (mat)[2][2] = 1.0; \
}

#define	mat3VecMul(dst,mat,v) \
{ \
	(dst)[0] = (mat)[0][0] * (v)[0] + (mat)[0][1] * (v)[1] + \
		(mat)[0][2] * (v)[2]; \
	(dst)[1] = (mat)[1][0] * (v)[0] + (mat)[1][1] * (v)[1] + \
		(mat)[1][2] * (v)[2]; \
	(dst)[2] = (mat)[2][0] * (v)[0] + (mat)[2][1] * (v)[1] + \
		(mat)[2][2] * (v)[2]; \
}

/*
 * Inserts three 3 - vectors into rows of 3x4 - matrix and sets last
 * column of matrix to zero
 */
#define	mat3x4SetRows(mat,v1,v2,v3) \
{ \
	(mat)[0][0] = (v1)[0]; (mat)[0][1] = (v1)[1]; (mat)[0][2] = (v1)[2]; \
		(mat)[0][3] = 0; \
	(mat)[1][0] = (v2)[0]; (mat)[1][1] = (v2)[1]; (mat)[1][2] = (v2)[2]; \
		(mat)[1][3] = 0; \
	(mat)[2][0] = (v3)[0]; (mat)[2][1] = (v3)[1]; (mat)[2][2] = (v3)[2]; \
		(mat)[2][3] = 0; \
}

/*
 * Misc
 */

#define	bboxClear(mins,maxs) \
{ \
	vec3Set(mins,MAXVEC,MAXVEC,MAXVEC); \
	vec3Set(maxs,-MAXVEC,-MAXVEC,-MAXVEC); \
}

#define	bboxAddPoint(mins,maxs,p) \
{ \
	if((mins)[0] > (p)[0]) (mins)[0] = (p)[0]; \
	if((maxs)[0] < (p)[0]) (maxs)[0] = (p)[0]; \
	if((mins)[1] > (p)[1]) (mins)[1] = (p)[1]; \
	if((maxs)[1] < (p)[1]) (maxs)[1] = (p)[1]; \
	if((mins)[2] > (p)[2]) (mins)[2] = (p)[2]; \
	if((maxs)[2] < (p)[2]) (maxs)[2] = (p)[2]; \
}

void	mat3Rotate(matrix3_t mat,vec3_t n,vec_t angle);
void	mat3x4FromPointDirAndUp(matrix3x4_t mat,vec3_t pt,vec3_t dir,vec3_t up);

#endif	/* !MVLIB_H_INLUDED */
